{-# LANGUAGE NoImplicitPrelude #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE DerivingVia #-}
{-# LANGUAGE TupleSections #-}
{-# OPTIONS_GHC -Wno-unrecognised-pragmas #-}
module Chart (newChart, FillerException(..), getUpdatedChartCache, getChunkedDay) where

import Import
import Control.Concurrent (forkIO)
import Control.Monad.Extra (ifM)
import Data.Aeson (eitherDecodeFileStrict')
import Data.FingerTree (FingerTree)
import Data.Time.Calendar (Day(..))
import Data.Time.Clock
import Data.Maybe (fromJust)
import RIO.FilePath
import RIO.List
import RIO.List.Partial

import qualified Data.FingerTree as FT
import qualified Data.HashMap.Strict as HM
import qualified Data.List as L
import qualified Data.Text as T
import qualified Data.Vector.Storable as VS

--import qualified Debug.Trace as D

data FillerException = QuitFiller
  deriving stock Show

instance Exception FillerException

newChart :: IBContract -> RIO App ()
newChart contract = do
  app <- ask
  let sym = (symbol :: IBContract -> Text) contract
      con = (conId :: IBContract -> Int) contract
      hmVar = appCharts . appRefs $ app
  hm <- liftIO . readTVarIO $ hmVar
  unless (sym `HM.member` hm) $ do
    today <- liftIO getCurrentDay :: RIO App Day
    (cacheErrors, cacheData) <- do
        files <- liftIO $ listDirectory $ "cache" </> show con
        res <- forM files $ \cacheFileName -> do
          let fname = "cache" </> show con </> cacheFileName
          ifM (liftIO $ doesFileExist fname)
              (bimap (\err -> cacheFileName <> ": "<> err) ((readMaybe $ dropExtension cacheFileName) :: Maybe Day,) <$> liftIO (eitherDecodeFileStrict' fname))
              (return $ Left $ "cachefile "<>cacheFileName<>" not found.")
        return $ partitionEithers res
    unless (null cacheErrors) $ forM_ cacheErrors $ \err -> logError (displayShow err)
    let (unknownDates, cacheData') = partition (isNothing . fst) cacheData
    unless (null unknownDates) $ logError $ display $ T.pack $ show (length unknownDates) <> " files had no parsable Date."
    let cacheData'' :: HashMap Day (FingerTree TimePoint ChartPoint) = HM.fromList $ bimap (fromMaybe today) FT.fromList <$> cacheData'
    logError $ displayShow $ HM.keys cacheData''
    c <- liftIO $ newTVarIO $ Chart con (fromMaybe FT.empty $ cacheData'' HM.!? today) (HM.delete today cacheData'') undefined defChartSettings defChartCacheSettings emptyChartCacheData Nothing False
    tid <- liftIO $ forkIO $ fillChart app contract c
    liftIO $ atomically $ do
      modifyTVar' c (\Chart{..} -> let fillerThread = tid in Chart{..})
      modifyTVar' hmVar (HM.insert sym c)

fillChart :: App -> IBContract -> TVar Chart -> IO ()
fillChart app contract cVar = runRIO app $ do
  let sym = (symbol :: IBContract -> Text) contract
  (tickerMapVar :: TVar (HashMap Int Text)) <- tickerIdToSymbol. appRefs <$> ask
  alreadyAdded <- (sym `elem`) . HM.elems <$> liftIO (readTVarIO tickerMapVar)
  unless alreadyAdded $ do
    tickerId <- (+1) . foldl' max 1000 . HM.keys <$> liftIO (readTVarIO tickerMapVar)
    let cancelSubscription = liftIO $ atomically $ do
         modifyTVar tickerMapVar (HM.delete tickerId)
         -- TODO: send cancel-request
    let sendQ = twsConnectionSend . twsConnectionRefs . appRefs $ app
    liftIO $ atomically $ do
      modifyTVar tickerMapVar (HM.insert tickerId sym)
      writeTQueue sendQ $ Msg_IB_OUT $ IB_RequestMktData tickerId contract "233" False False
    handle (\QuitFiller -> cancelSubscription >> exitSuccess) $
      forever $ do
        -- chart dirty? set clean & begin work
        c <- liftIO (readTVarIO cVar)
        when (chartDirty c) $ do
          liftIO $ atomically $ modifyTVar cVar (\c' -> c' { chartDirty = False })
          (lUpdate, cachePoints) <- getUpdatedChartCache c Nothing
          liftIO $ atomically $ modifyTVar cVar (\c' -> c' { chartCache = cachePoints, lastCacheUpdate = lUpdate })
          return ()
        threadDelay 1000000 -- sleep 5 seconds

chunkChart :: Int -> Int -> Int -> FingerTree TimePoint ChartPoint -> [(TimePoint,[ChartPoint])]
chunkChart from until range tree = go from range interval
    where
      lastItem = case FT.viewr interval of
                  FT.EmptyR -> until
                  (_ FT.:> ChartPoint{..}) -> (\(TimePoint x) -> x) timeOfDay
      interval = FT.takeUntil (\(TimePoint x) -> x > until)
               . FT.dropUntil (\(TimePoint x) -> x > from)
               $ tree
      go f i t
        | f+i >= lastItem = [(TimePoint (f+i), toList t)]
        | otherwise       = let (a, b) = FT.split (\(TimePoint x) -> x > f+i) t
                            in (TimePoint (f+i),toList a) : go (f+i) i b

toCachePoint :: (TimePoint,[ChartPoint]) -> ChartPoint
toCachePoint (t,[]) = ChartPoint t (-1) 0 []
toCachePoint (t,as) = ChartPoint t m vol [OLHC o l h c]
  where
    vol = sum $ pointVolume <$> as
    as' = pointValue <$> as
    ms' = sum $ (\x -> pointValue x * pointVolume x) <$> as
    m = ms' / vol
    o = head as'
    c = last as'
    l = minimum as'
    h = maximum as'

getUpdatedChartCache :: MonadIO m => Chart -> Maybe ChartCacheSettings -> m (Maybe TimePoint, ChartCacheData)
getUpdatedChartCache Chart{..} chartCacheSettings' = do
    let ChartCacheSettings cRes cTicks = fromMaybe chartCacheSettings chartCacheSettings'
    now <- liftIO getCurrentTime
    -- - recalculate cacheUpdateStart & cacheUpdateEnd
    let cacheUpdateEnd   = (utcTimeToSeconds now + cRes) `div` cRes
        cacheUpdateStart = cacheUpdateEnd - (cRes * cTicks)
        ChartCacheData ccData ccRange ccFill = chartCache
        chunks = case ccFill of
                  Nothing -> [(cacheUpdateStart, cacheUpdateEnd)]
                  Just (mi, ma) -> if
                        -- Interval mi-ma already updated. get remaining intervals
                        -- mi  ma  cUS cUE
                      | ma < cacheUpdateStart -> [(cacheUpdateStart, cacheUpdateEnd)]
                        -- mi?  cUS  mi?  ma  cUE -> [ma,cUE] + rest
                      | ma < cacheUpdateEnd -> [(ma,cacheUpdateEnd)] <> if
                          -- cUS  mi  ma
                          | mi > cacheUpdateStart -> [(cacheUpdateStart, mi)]
                          | otherwise -> []
                      | otherwise -> []
        -- - chunk them with chunhChart
        chunkedChart = L.concat $ for chunks $ \(start, end) -> chunkChart (min 0 start) end cRes chartData
        lUpdate = cacheUpdateEnd - cRes
        -- - calculate Range and need for shift
        shiftNeccessary = case ccRange of
                            Nothing     -> False
                            Just (_,ma) -> cacheUpdateEnd > ma
        (newRange, shiftInterval) = if not shiftNeccessary && isJust ccRange then (fromJust ccRange, 0) else 
                      let interval = (cTicks + 20) `div` 20 :: Int
                          iTo = (cacheUpdateEnd + cRes * interval) `div` (cRes * interval)
                          iFrom = iTo - 20 * (cRes * interval)
                      in ((iFrom,iTo), interval)

    -- create data for updates
    let timePointToIndex' :: (Int, Int) -> (Int, Int) -> TimePoint -> Int
        timePointToIndex' (cRes', cTicks') (rFrom, rTo) (TimePoint p) = let result = (p - rFrom) `div` cRes in
                            if
                              | p < rFrom || p > rTo            -> error $ "timePointToIndex' " <> show ((cRes', cTicks'),(rFrom, rTo),TimePoint p) <> ": " <> show p <> " out of range."
                              | result < 0 || result >= cTicks' -> error $ "timePointToIndex' " <> show ((cRes', cTicks'),(rFrom, rTo),TimePoint p) <> ": " <> show result <> " is invalid index"
                              | otherwise                       -> result
        timePointToIndex :: TimePoint -> Int
        timePointToIndex = timePointToIndex' newRange (cRes, cTicks)
        -- fold instead of map, so we can carry over cached calculatinos from element to element (for i.e. ema/sma)
        (!foldedData,_,_) = foldl' folder ([], ChartPoint 0 0 0 [], mempty) chunkedChart
        folder :: ([(Int, ChartPoint)], ChartPoint, HM.HashMap Int [Float]) -> (TimePoint, [ChartPoint]) -> ([(Int, ChartPoint)], ChartPoint, HM.HashMap Int [Float])
        folder old (_,[]) = old
        folder (acc, lastPoint, smaMap) (tp,cdata) = ((timePointToIndex tp, cp):acc, cp, foldl' (.) id smaUpdates smaMap)
            where
              cp = ChartPoint tp m vol studies
              (studies, smaUpdates) = unzip $ catMaybes $ (chartStudySettings chartSettings :: [ChartStudyType]) <&> \case
                  ChartStudyTypeOpen   -> Just (OLHC o l h c, id)
                  ChartStudyTypeHigh   -> Nothing
                  ChartStudyTypeLow    -> Nothing
                  ChartStudyTypeClose  -> Nothing
                  ChartStudyTypeDirect -> Nothing
                  ChartStudyTypeSMA w  -> Nothing --TODO: fixme #13, do the calculation & use data inside smaMap and return (SMA x y, smaUpdate)
                  ChartStudyTypeVolume -> Just (Volume vol, id)
              vol = sum $ pointVolume <$> cdata
              as' = pointValue <$> cdata
              ms' = sum $ (\x -> pointValue x * pointVolume x) <$> cdata
              m = ms' / vol
              o = head as'
              c = last as'
              l = minimum as'
              h = maximum as'
        -- plan the actual work
        updates = chartStudySettings chartSettings <&> \cs ->
            -- check if thing is in hashmap
            let vec = case ccData HM.!? cs of
                    Nothing -> VS.replicate cTicks 0
                    Just a  -> a
                -- shift if neccessary
                vec' = if shiftNeccessary && cs `HM.member` ccData then
                          -- TODO: unsafeSlice && unsafeUpdate_ - see #16
                          let sliceLength = VS.length vec - shiftInterval
                          in VS.update_ vec (VS.enumFromN 0 sliceLength) (VS.slice shiftInterval sliceLength vec)
                       else
                        vec
            -- NOW:
            --  newRange is (rangeFrom, rangeTo) in seconds with 0 being last midnight
            --  vec' has that range mapped to indices [0,cTicks-1] with every cRes being one time-slice
            --  everything in the chunkedChart-list has to be inserted if in range using the same time-notation like the range
            --  (hint: chunkedChart is (time, [data]), with time being the point at the end of each contained interval)
                findOLHC   as = fromJust $ find (\case OLHC{}   -> True; _ -> False) as
                findVolume as = fromJust $ find (\case Volume{} -> True; _ -> False) as
            in case cs of
              ChartStudyTypeDirect -> HM.insert cs (vec' VS.// ((\(i, ChartPoint _ x _ _) -> (i,x))                       <$> foldedData))
              ChartStudyTypeOpen   -> HM.insert cs (vec' VS.// ((\(i, ChartPoint _ _ _ s) -> (i,olhc_open  $ findOLHC s)) <$> foldedData))
              ChartStudyTypeLow    -> HM.insert cs (vec' VS.// ((\(i, ChartPoint _ _ _ s) -> (i,olhc_low   $ findOLHC s)) <$> foldedData))
              ChartStudyTypeHigh   -> HM.insert cs (vec' VS.// ((\(i, ChartPoint _ _ _ s) -> (i,olhc_high  $ findOLHC s)) <$> foldedData))
              ChartStudyTypeClose  -> HM.insert cs (vec' VS.// ((\(i, ChartPoint _ _ _ s) -> (i,olhc_close $ findOLHC s)) <$> foldedData))
              ChartStudyTypeSMA x  -> id --TODO: implement #13
              ChartStudyTypeVolume -> HM.insert cs (vec' VS.// ((\(i, ChartPoint _ _ _ s) -> (i,volume $ findVolume s))   <$> foldedData))
    return (Just $ TimePoint lUpdate, ChartCacheData (foldl' (.) id updates ccData) (Just newRange) Nothing) -- FIXME: fillData still missing

getChunkedDay :: Chart -> Maybe Int -> (Maybe TimePoint, [ChartPoint])
getChunkedDay Chart{..} chunkResolution = (lUpdate, toCachePoint <$> chunkedData)
    where
        chunkedData = chunkChart 0 86400 (fromMaybe (chartCacheResolution chartCacheSettings) chunkResolution) chartData
        lUpdate     = fmap fst . lastMaybe $ chunkedData
